# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class Template2LevMod(Template):       
    'Conducts regression imputation.'

    __version__ = '1.0.0'

    tags = [ 'Model', 'Normal', 'Data manipulation', 'Missing data']
    engines = ['Python_script']

    inputs = '''
variable_with_missing = DataVector('Response (variable with missing values): ')
x = DataMatrix('Explanatory variables you want to use to predict missing values (note: no need to specify intercept): ', allow_cat = True)
stochastic = Boolean("Do you want to perform stochastic (random) regression imputation? (If answer 'no', then regression mean imputation is instead performed): ")
if stochastic:
    imputation_seed = Integer('What seed do you want to use for the random number generator?')
variable_with_regression_imputation = Text('Name of new variable with imputed values: ')
new_dataset = Text('Name of output dataset: ')
'''

    pythonscript = '''
## Below works, but some redundancy, etc., to tidy up
## Plus, could poss add a stacked histogram: e.g. this might help: http://stackoverflow.com/questions/18449602/matplotlib-creating-stacked-histogram-from-three-unequal-length-arrays

from EStat.stats.utils.kernal import Kernal

import numpy
import scipy
import scipy.stats

import base64

from io import BytesIO

from matplotlib.figure import Figure
import matplotlib.lines as lines
from matplotlib.backends.backend_agg import FigureCanvasAgg

import EStat
from EStat.Templating import *
from EStat.DTAFile import DTAFile


def stars(pval):
    if pval > 0.05:
        return '.'
    else:
        s = '*'
        if pval < 0.01:
            s += '*'
        if pval < 0.001:
            s += '*'
        return s

def pval2str(pval):
    import numpy
    if pval < 0.001:
        return '< 0.001'
    else:
        return str(numpy.round(pval, 3))

def roundtosf(num, sf):
    import numpy
    return numpy.round(num, sf-int(numpy.trunc(numpy.log10(numpy.abs(num)))))

xnames = list(x)
xorder = [1 for i in range(0, len(x))]

missing_indicator = numpy.array(numpy.ma.getmaskarray(datafile.variables[str(variable_with_missing)]['data']), dtype=int)
Y = datafile.variables[variable_with_missing]['data'][missing_indicator == 0]
X = numpy.ma.column_stack([datafile.variables[i]['data'][missing_indicator == 0] for i in x])

X = numpy.ma.column_stack((X, numpy.ma.ones(len(Y))))
xnames.append('_cons')
xorder.append(0)

betamat, residuals, rank, s = numpy.linalg.lstsq(X, Y)

if stochastic:
    fitted_orig = numpy.dot(betamat, X.T)
    yres = Y - fitted_orig
    sse = sum(yres**2)
    sigmasq = sse / (len(Y) - len(xnames))
    sigma = numpy.sqrt(sigmasq)
    numpy.random.seed(int(imputation_seed))

FULL_X = numpy.ma.column_stack([datafile.variables[i]['data'] for i in x])
FULL_X = numpy.ma.column_stack((FULL_X, numpy.ma.ones(datafile.nobs)))

fitted = numpy.dot(betamat, FULL_X.T)

IMP_Y = numpy.ma.array(datafile.variables[variable_with_missing]['data'])

if stochastic:
    IMP_Y[missing_indicator == 1] = numpy.random.normal(fitted[missing_indicator == 1], sigma, numpy.sum(missing_indicator))
    graph_title = 'Exploring Stochastic (Random) Regression Imputation'
else:
    IMP_Y[missing_indicator == 1] = fitted[missing_indicator == 1]
    graph_title = 'Exploring Regression (Conditional) Mean Imputation'

## Exporting dataset with imputed variable

retval = DTAFile()

retval.nobs = datafile.nobs
for k in datafile.variables.keys():
    retval.addvariable(k, data = datafile.variables[k]['data'])

retval.addvariable(str(variable_with_regression_imputation), data = IMP_Y)
outputs[str(new_dataset)] = retval

## Producing densityplot

fig = Figure(figsize=(8,8)) # Units are inches
ax = fig.add_subplot(111, xlabel = variable_with_missing, ylabel = 'kernel density')

ykernal=numpy.zeros(1000)
xkernal=numpy.zeros(1000)

Kernal(Y.compressed(), ykernal, xkernal, 1.0, 0, 1, 0, -1)
ax.plot(xkernal, ykernal, label = "Observed variable")
Kernal(IMP_Y.compressed(), ykernal, xkernal, 1.0, 0, 1, 0, -1)
ax.plot(xkernal, ykernal, label = "With imputed values")

ax.tick_params(axis='both', which='major', labelsize=10)

leg=ax.legend(prop={'size':10})
###request semi-transparent legend (in case of any overlap):
leg.get_frame().set_alpha(0.5)
#leg.set_title(variable_with_missing)

fig.suptitle(graph_title)

canvas = FigureCanvasAgg(fig)

buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')
buf.seek(0)
outputs['densityplot.svg'] = ImageOutput(buf.getvalue())
buf.close()

###################################################################

## Exploring other densityplots

fig = Figure(figsize=(8,8)) # Units are inches
ax = fig.add_subplot(111, xlabel = variable_with_missing, ylabel = 'kernel density')

fig_separate = Figure(figsize=(8,8)) # Units are inches

ykernal=numpy.zeros(1000)
xkernal=numpy.zeros(1000)

ax_separate = fig_separate.add_subplot(2, 1 , 1, xlabel = variable_with_missing + ': observed values only', ylabel = 'kernel density')
Kernal(Y.compressed(), ykernal, xkernal, 1.0, 0, 1, 0, -1)

ax_separate.plot(xkernal, ykernal, 'b')

Kernal(Y.compressed(), ykernal, xkernal, 1.0, 0, 1, 0, -1)
ax.plot(xkernal, ykernal, label = 'Observed values only')

yvals = IMP_Y
filter = missing_indicator

Kernal(yvals[filter == 1].compressed(), ykernal, xkernal, 1.0, 0, 1, 0, -1)
ax.plot(xkernal, ykernal, label = 'Imputed values only')
ax_separate = fig_separate.add_subplot(2, 1, 2, xlabel = variable_with_missing + ': imputed values only', ylabel='kernel density')
ax_separate.plot(xkernal, ykernal, 'g')

ax.tick_params(axis='both', which='major', labelsize=10)

leg=ax.legend(prop={'size':10})
#request semi-transparent legend (in case of any overlap):
leg.get_frame().set_alpha(0.5)

fig.suptitle(graph_title)
fig_separate.suptitle(graph_title)

canvas = FigureCanvasAgg(fig)

buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')
buf.seek(0)
outputs['densityplot_only.svg'] = ImageOutput(buf.getvalue())
buf.close()

fig_separate.tight_layout(rect=[0, 0.02, 1, 0.98])
canvas = FigureCanvasAgg(fig_separate)

buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')
buf.seek(0)
outputs['densityplot_only_panel.svg'] = ImageOutput(buf.getvalue())
buf.close()
###################################################################

## Producing boxplot

vals = []
#xdata = datafile.variables[variable]['data']
#filterval = datafile.variables[group_by]['data']
#labvals = numpy.sort(numpy.unique(filterval.compressed()))

#for i in labvals:
#    filt = filterval == i
#    vals.append(xdata[filt].compressed())

vals.append(Y.compressed())
vals.append(IMP_Y.compressed())

fig_box = Figure(figsize=(8,8))

#ax = fig.add_subplot(100 + 10 + 1, xlabel = group_by, ylabel = variable)
ax = fig_box.add_subplot(100 + 10 + 1, ylabel = variable_with_missing)
vertical = True

ax.boxplot(vals, labels = (['Observed variable', 'With imputed values']), vert = vertical)

fig_box.suptitle(graph_title)

fig_box.tight_layout(rect=[0, 0.02, 1, 0.98])

canvas = FigureCanvasAgg(fig_box)
buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['box.svg'] = ImageOutput(buf.getvalue())
buf.close()

## Producing plot by rank

fig_rank = Figure(figsize=(8,8))
ax = fig_rank.add_subplot(100 + 10 + 1, xlabel = "Rank", ylabel = variable_with_missing)

vals = IMP_Y
ranks = IMP_Y.argsort().argsort()

filter = missing_indicator

ax.plot(ranks[filter == 0], vals[filter == 0], 'o', fillstyle='none', markersize = '4', color="green", label = "Observed value")
ax.plot(ranks[filter == 1], vals[filter == 1], 'o', fillstyle='none', markersize = '4', color="red", label = "Imputed value")

#box = ax.get_position()
#ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

#leg=ax.legend(prop={'size':10})
#leg=ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), prop={'size':10})
leg=ax.legend(loc='upper left', prop={'size':10}, numpoints = 1)
###request semi-transparent legend (in case of any overlap):
leg.get_frame().set_alpha(0.5)
#leg.set_title(variable_with_missing)

fig_rank.suptitle(graph_title)

canvas = FigureCanvasAgg(fig_rank)
buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['rank.svg'] = ImageOutput(buf.getvalue())
buf.close()

## Producing scatterplot by x

for i in range(0, len(x)):
    fig = Figure(figsize=(8,8))
    ax = fig.add_subplot(100 + 10 + 1, xlabel = x[i], ylabel = variable_with_missing)

    yvals = IMP_Y
    xvals = datafile.variables[x[i]]['data']

    filter = missing_indicator

    ax.plot(xvals[filter == 0], yvals[filter == 0], 'o', fillstyle='none', markersize = '4', color="green", label = "Observed value")
    ax.plot(xvals[filter == 1], yvals[filter == 1], 'o', fillstyle='none', markersize = '4', color="red", label = "Imputed value")

    #box = ax.get_position()
    #ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    #leg=ax.legend(prop={'size':10})
    #leg=ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), prop={'size':10})
    leg=ax.legend(loc='upper left', prop={'size':10}, numpoints = 1)
    ###request semi-transparent legend (in case of any overlap):
    leg.get_frame().set_alpha(0.5)
    #leg.set_title(variable_with_missing)

    fig.suptitle(graph_title)

    canvas = FigureCanvasAgg(fig)
    buf = BytesIO()
    canvas.print_figure(buf, dpi=80, format='svg')

    buf.seek(0)
    outputs['fig_scatter_' + x[i] + '.svg'] = ImageOutput(buf.getvalue())
    buf.close()

## Producing table of means/sds of original vs imputed variable

tabout = TabularOutput()
headings = [str(variable_with_missing)]
headings.append('Mean')
headings.append('SD')
headings.append('N')
tabout.column_headings = headings
row = []
row.append(roundtosf(numpy.mean(Y),3))
row.append(roundtosf(numpy.std(Y),3))
row.append(sum(1 - missing_indicator))
tabout.add_row('Observed variable', row)
row = []
row.append(roundtosf(numpy.mean(IMP_Y),3))
row.append(roundtosf(numpy.std(IMP_Y),3))
row.append(len(missing_indicator))
tabout.add_row('With imputed values', row)
row = []
row.append(roundtosf(numpy.mean(IMP_Y[filter == 1]),3))
row.append(roundtosf(numpy.std(IMP_Y[filter == 1]),3))
row.append(len(IMP_Y[filter == 1]))
tabout.add_row('Imputed values only', row)

outputs['table'] = tabout

'''
